use io;

/* Read instance data. The input files follow the "longjianyu" format */
function input() {
    usage = "Usage: localsolver pcvrp.lsp inFileName=inputFile "
        + "[solFileName=outputFile] [lsTimeLimit=timeLimit]";

    if (inFileName == nil) throw usage;
    
    readInputPcvrp();

    computeDistanceMatrix();
}

/* Declare the optimization model */
function model() {
    // Sequence of customers visited by each truck
    customersSequences[k in 0...nbTrucks] <- list(nbCustomers);

    // A customer might be visited by only one truck
    constraint disjoint[k in 0...nbTrucks](customersSequences[k]);

    for [k in 0...nbTrucks] {
        local sequence <- customersSequences[k];
        local c <- count(sequence);

        // A truck is used if it visits at least one customer
        trucksUsed[k] <- c > 0;

        // The quantity needed in each route must not exceed the truck capacity
        routeQuantities[k] <- sum(sequence, j => demands[j]);
        constraint routeQuantities[k] <= truckCapacity;

        // Distance traveled by truck k
        routeDistances[k] <- sum(1...c, i => distanceMatrix[sequence[i - 1]][sequence[i]])
                + (c > 0 ? (distanceDepot[sequence[0]] + distanceDepot[sequence[c - 1]]) : 0);

        // Route prize of truck k
        routePrizes[k] <- sum(sequence, j => prizes[j]);
    }

    // Total nb demands satisfied
    totalQuantity <- sum[k in 0...nbTrucks](routeQuantities[k]);
    
    // Minimal number of demands to satisfy
    constraint totalQuantity >= demandsToSatisfy;

    // Total nb trucks used
    nbTrucksUsed <- sum[k in 0...nbTrucks](trucksUsed[k]);

    // Total distance traveled
    totalDistance <- sum[k in 0...nbTrucks](routeDistances[k]);

    // Total prize
    totalPrize <- sum[k in 0...nbTrucks](routePrizes[k]);

    // Objective: minimize the number of trucks used, then maximize the total prize and minimize the distance traveled
    minimize nbTrucksUsed;
    maximize totalPrize;
    minimize totalDistance;
}

/* Parametrize the solver */
function param() { 
    if (lsTimeLimit == nil) lsTimeLimit = 20;
}

/* Write the solution in a file with the following format:
 * - total prize, number of trucks used and total distance
 * - for each truck the customers visited (omitting the start/end at the depot)
 * - number of unvisited customers, demands satisfied */
function output() {
    if (solFileName == nil) return;
    local outfile = io.openWrite(solFileName);

    outfile.println(totalPrize.value, " ", nbTrucksUsed.value, " ", totalDistance.value);
    nbUnvisitedCustomers = nbCustomers;

    for [k in 0...nbTrucks] {
        if (trucksUsed[k].value != 1) continue;
        // Values in sequence are in [0...nbCustomers].
        // +1 is to put it back in [1...nbCustomers+1]
        // as in the data files (0 being the depot)
        for [customer in customersSequences[k].value] {
            outfile.print(customer + 1, " ");
            nbUnvisitedCustomers -= 1;
        }
        outfile.println();
    }

    outfile.println(nbUnvisitedCustomers, " ", totalQuantity.value);
    
    outfile.close();
}

function readInputPcvrp() {
    local inFile = io.openRead(inFileName);

    nbTrucks = inFile.readInt();
    truckCapacity = inFile.readInt();
    demandsToSatisfy = inFile.readInt();

    n = 0;
    while (!inFile.eof()) {
        if (n != inFile.readInt()) throw "Unexpected index";
        customersX[n] = round(inFile.readDouble());
        customersY[n] = round(inFile.readDouble());
        if (n == 0) {
            depotId = n;
        } else {
            // -1 because depot has no demand neither prize
            demands[n - 1] = inFile.readInt();
            prizes[n - 1] = inFile.readInt();
        }
        n += 1;
    }

    nbCustomers = n - 1;

    inFile.close();
}

/* Compute the distance matrix */
function computeDistanceMatrix() {
    for [i in 0...nbCustomers] {
        distanceMatrix[i][i] = 0;
        for [j in i+1...nbCustomers] {
            local localDistance = computeDist(i + 1, j + 1);
            distanceMatrix[j][i] = localDistance;
            distanceMatrix[i][j] = localDistance;
        }
    }

    for [i in 0...nbCustomers] {
        local localDistance = computeDist(0, i + 1);
        distanceDepot[i] = localDistance;
    }
}

/* Compute the euclidean distance */
function computeDist(i, j) {
    local exactDist = sqrt(pow((customersX[i] - customersX[j]), 2)
        + pow((customersY[i] - customersY[j]), 2));
    return round(exactDist);
}
